gLevelValues.exitCastleLevel = 16
gLevelValues.exitCastleArea = 1
gLevelValues.exitCastleWarpNode = 1

gBehaviorValues.KingBobombFVel = 3.0
gBehaviorValues.KingBobombYawVel = 256
gBehaviorValues.KingBobombHealth = 3

gBehaviorValues.KingWhompHealth = 3
gBehaviorValues.KoopaThiAgility = 6.0
gBehaviorValues.KoopaBobAgility = 4.0

gBehaviorValues.MipsStar1Requirement = 20
gBehaviorValues.MipsStar2Requirement = 60

gBehaviorValues.ToadStar1Requirement = 15
gBehaviorValues.ToadStar2Requirement = 25
gBehaviorValues.ToadStar3Requirement = 35

gLevelValues.pssSlideStarTime = 630

gLevelValues.metalCapDuration = 600
gLevelValues.wingCapDuration = 1800
gLevelValues.vanishCapDuration = 600

gLevelValues.metalCapDurationCotmc = 1
gLevelValues.wingCapDurationTotwc = 1
gLevelValues.vanishCapDurationVcutm = 1

gBehaviorValues.ChillBullyDeathPosY = -3400

hook_event(HOOK_ON_PAUSE_EXIT, function(exitToCastle)
    if not exitToCastle and
    gNetworkPlayers[0].currLevelNum == LEVEL_TTM then
       return false
    end
 end)

-- Faster hangable ceiling --
-- Original Code by Sunk

---@param num number
---@return number
local function convert_s16(num)
    num = num & 0xFFFF
    return ((num >= 0x7FFF) and (num - 0x10000) or num)
end

---@param m MarioState
local function update_custom_hang_moving(m)
    local stepResult = 0
    local nextPos = {}
    local maxSpeed = 26

    m.forwardVel = m.forwardVel + 16
    if m.forwardVel > maxSpeed then
        m.forwardVel = maxSpeed
    end

    m.faceAngle.y = m.intendedYaw - approach_s32(convert_s16(m.intendedYaw - m.faceAngle.y), 0, 0x800, 0x800)

    m.slideYaw = m.faceAngle.y
    m.slideVelX = m.forwardVel * sins(m.faceAngle.y)
    m.slideVelZ = m.forwardVel * coss(m.faceAngle.y)

    m.vel.x = m.slideVelX
    m.vel.y = 0.0
    m.vel.z = m.slideVelZ

    nextPos.x = m.pos.x - m.ceil.normal.y * m.vel.x
    nextPos.z = m.pos.z - m.ceil.normal.y * m.vel.z
    nextPos.y = m.pos.y

    stepResult = perform_hanging_step(m, nextPos)

    vec3f_copy(m.marioObj.header.gfx.pos, m.pos)
    vec3s_set(m.marioObj.header.gfx.angle, 0, m.faceAngle.y, 0)
    return stepResult
end

---@param m MarioState
function act_custom_hang_moving(m)
    if m.input & INPUT_A_DOWN == 0 then
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    if m.input & INPUT_Z_PRESSED ~= 0 then
        return set_mario_action(m, ACT_GROUND_POUND, 0)
    end

    if m.ceil == nil or m.ceil.type ~= SURFACE_HANGABLE then
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    if m.actionArg & 1 ~= 0 then
        set_mario_animation(m, MARIO_ANIM_MOVE_ON_WIRE_NET_RIGHT)
    else
        set_mario_animation(m, MARIO_ANIM_MOVE_ON_WIRE_NET_LEFT)
    end

    if m.marioObj.header.gfx.animInfo.animFrame == 12 then
        play_sound(SOUND_ACTION_HANGING_STEP, m.marioObj.header.gfx.cameraToObject)
        queue_rumble_data_mario(m, 5, 30)
    end

    if is_anim_past_end(m) ~= 0 then
        m.actionArg = m.actionArg ~ 1
        if m.input & INPUT_ZERO_MOVEMENT ~= 0 then
            return set_mario_action(m, ACT_HANGING, m.actionArg)
        end
    end

    if update_custom_hang_moving(m) == 2 --[[HANG_LEFT_CEIL]] then
        set_mario_action(m, ACT_FREEFALL, 0)
    end

    return 0
end

hook_mario_action(ACT_HANG_MOVING, act_custom_hang_moving)

-- Faster Swimming --

function mario_before_phys_step(m)
    local hScale = 1.0
    local vScale = 1.0

    if (m.action & ACT_FLAG_SWIMMING) ~= 0 then
        hScale = hScale * 1.5
        if m.action ~= ACT_WATER_PLUNGE then
            vScale = vScale * 1.5
        end
    end

    m.vel.x = m.vel.x * hScale
    m.vel.y = m.vel.y * vScale
    m.vel.z = m.vel.z * hScale
end

hook_event(HOOK_BEFORE_PHYS_STEP, mario_before_phys_step)

-- Elevator Fix --

local jumpActions = {
    [ACT_JUMP] = true,
    [ACT_DOUBLE_JUMP] = true,
    [ACT_TRIPLE_JUMP] = true,
    [ACT_SPECIAL_TRIPLE_JUMP] = true,
    [ACT_STEEP_JUMP] = true,
    [ACT_SIDE_FLIP] = true,
    [ACT_BACKFLIP] = true,
    [ACT_LONG_JUMP] = true,
    [ACT_RIDING_SHELL_JUMP] = true
}

local function elevator_fix(m)
    if jumpActions[m.action] then
        if m.pos.y < m.floorHeight - 6 then
            m.pos.y = m.pos.y + 5
            m.vel.y = m.vel.y + 11
        end
    end
end

-- Leave Level Stars --

---@param m MarioState
---@param np NetworkPlayer
local function leave_level(m, np)
    if (np.currLevelNum == LEVEL_SL and np.currAreaIndex == 3) or 
       (np.currLevelNum == LEVEL_BITDW and np.currAreaIndex == 2) or 
       (np.currLevelNum == LEVEL_RR) or 
       (np.currLevelNum == LEVEL_BITS and np.currAreaIndex == 2) then
        if m.action == ACT_STAR_DANCE_NO_EXIT then
            m.actionArg = 0
        end
    end

    gLevelValues.coinsRequiredForCoinStar = (np.currLevelNum == LEVEL_SL or np.currLevelNum == LEVEL_RR) and 0 or 100
end

---@param m MarioState
---@param np NetworkPlayer
local function force1hp(m, np)
    if (np.currLevelNum == LEVEL_RR) or (np.currLevelNum == LEVEL_BOB and np.currAreaIndex == 2) then
        if m.healCounter > 12 then
            m.healCounter = 0
        end
    end
end

local perfectTimer = 0

local incrementActions = {
    [ACT_DIVE_SLIDE] = true,
    [ACT_AIR_HIT_WALL] = true,
    [ACT_SOFT_BONK] = true
}

local perfectActions = {
    [ACT_WALL_KICK_AIR] = true,
    [ACT_FORWARD_ROLLOUT] = true,
    [ACT_BACKWARD_ROLLOUT] = true
}

---@param m MarioState
local function update_perfect_timer(m)
    if m.playerIndex ~= 0 then return end

    if incrementActions[m.action] then
        if perfectTimer < 3 then
            perfectTimer = perfectTimer + 1
        end
    else
        perfectTimer = 0
    end
end

---@param m MarioState
local function mario_on_set_action(m)
    if m.playerIndex == 0 then
        if (perfectActions[m.action] and perfectTimer == 1) or
           (m.action == ACT_JUMP_KICK and m.prevAction == ACT_MOVE_PUNCHING and m.forwardVel > 20) then
            m.actionArg = 99
        end
    end

    if (perfectActions[m.action] and m.actionArg == 99) or 
       (m.action == ACT_JUMP_KICK and m.actionArg == 99) then
        m.particleFlags = m.particleFlags | PARTICLE_SPARKLES
    end

    if m.playerIndex ~= 0 then return end

    elevator_fix(m)
end

---@param m MarioState
local function mario_update(m)
    if m.playerIndex ~= 0 then return end

    ---@type NetworkPlayer
    local np = gNetworkPlayers[0]

    if m.numLives < 100 then
        m.numLives = 100
    end

    if m.action == ACT_EMERGE_FROM_PIPE and m.actionTimer <= 1 and np.currLevelNum == LEVEL_SSL then
        m.pos.y = m.pos.y + 50
    end

    force1hp(m, np)
    leave_level(m, np)
    update_perfect_timer(m)
end

hook_event(HOOK_ON_SET_MARIO_ACTION, mario_on_set_action)
hook_event(HOOK_MARIO_UPDATE, mario_update)